/***************************************************************************

			Adlib Interface for Gameboy 97 Sound version 2

 **************************************************************************/

#include <stdio.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned int WORD;

FILE *f;

static WORD Freq[] = {					/* Frequency table from FMFREQS.H */
#include "FMFreqs.h"
};


/***************** 	MACHINE CODE IMPLEMENTATION	 **********************/

#define ADLIB	(0x388)

static int Vol[] = { 0,32,34,36,38,40,42,44,46,48,50,52,54,56,58,60 };

static BYTE InstAd[] = { 0,1,2,8 };		/* Base address of the channels */

/************************************************************************/
/*																		*/
/*					Write a byte to the adlib port 						*/
/*																		*/
/************************************************************************/

void WriteFM(BYTE Register,BYTE Data)
{
BYTE i,n;
n = 0;
outportb(ADLIB,Register);				/* Output the address */
for (i = 0;i < 28;i++) n++;				/* Short Delay */
outportb(ADLIB+1,Data);					/* Output the Data */
for (i = 0;i < 68;i++) n++;				/* Short Delay */
}

/************************************************************************/
/*																		*/
/*							Read the Status								*/
/*																		*/
/************************************************************************/

BYTE ReadFM(void)
{
return(inportb(ADLIB));
}

/************************************************************************/
/*																		*/
/*					Detect existence of ADLIB Card						*/
/*																		*/
/************************************************************************/

BYTE DetectAdlibCard(void)
{
BYTE Value,Ok,i;
Ok = 1;
WriteFM(1,0);   						/* Disable test register */
WriteFM(4,0x60);						/* Disable connection/stop counters */
WriteFM(4,0x80);						/* Delete Status Register */
if ((ReadFM() & 0xE0) != 0) Ok = 0;   	/* Must be 0 in high bits */
WriteFM(2,0xFF);						/* Counter 1 = 255 */
WriteFM(4,0x21);						/* Connection on & start */
for (i = 0;i < 200;i++) Value=ReadFM();	/* Read port 200 times */
if ((ReadFM() & 0xE0) != 0xC0) Ok = 0;	/* Counter should have timed out */
WriteFM(4,0x60);						/* Disable connection/counters */
WriteFM(4,0x80);
WriteFM(8,0);							/* Melody mode on */
WriteFM(0xBD,0);

/* Now we need to write the data at AdlibSetupData to the Adlib card */

return(Ok);
}

/************************************************************************/
/*																		*/
/*							Set Volume on Channel						*/
/*																		*/
/************************************************************************/

void SetVolume(BYTE Channel,BYTE Volume)
{

if (Volume > 15) Volume = 15;			/* Check volume level */
WriteFM(0x43+InstAd[Channel],63-Vol[Volume]);/* Update the volume */
}

/************************************************************************/
/*																		*/
/*						Set Frequency on Channel						*/
/*																		*/
/************************************************************************/

void SetFrequency(BYTE Channel,WORD Frequency)
{
BYTE Instr,FreqLo,FreqHi;

FreqLo = (Freq[Frequency*2+1] & 0xFF);  /* This translation is saved */
FreqHi = (Freq[Frequency*2+1] >> 8) +	/* so just need to write the data */
					(Freq[Frequency*2] << 2) + 32;
WriteFM(0xA0+Channel,FreqLo);
WriteFM(0xB0+Channel,FreqHi);
fprintf(f,"db %u,%u\n",FreqLo,FreqHi);	/* Output frequency data for later */
}

/************************************************************************/

/************************************************************************/
/*																		*/
/*								Set Voices								*/
/*																		*/
/************************************************************************/

#define XWriteFM(d,a) { WriteFM(a,d);fprintf(f," db %u,%u\n",a,(d) & 0xFF); }

static BYTE Sound_Melodic[11]=
{
  0x23,0x21,    /* AM_VIB_EG-TYP_KSR_MULTI */
  0x1E,0x00,    /* KSL_TL                  */
  0xFF,0xFF,    /* AR_DR                   */
  0x08,0x08,    /* SL_RR                   */
  0x00,0x00,    /* WS                      */
  0x0E          /* FB_FM                   */
};

static BYTE Sound_Wave[11]=
{
  0x23,0x21,    /* AM_VIB_EG-TYP_KSR_MULTI */
  0x1E,0x00,    /* KSL_TL                  */
  0xFF,0xFF,    /* AR_DR                   */
  0x08,0x08,    /* SL_RR                   */
  0x00,0x00,    /* WS                      */
  0x0A          /* FB_FM                   */
};

static BYTE Sound_Noise[11]=
{
  0x2E,0x2E,    /* AM_VIB_EG-TYP_KSR_MULTI */
  0x40,0x03,    /* KSL_TL                  */
  0xFF,0xFF,    /* AR_DR                   */
	0x0F,0x0F,    /* SL_RR                   */
  0x00,0x03,    /* WS                      */
  0x0E          /* FB_FM                   */
};

static void Adlib_SetVoice (char *voice,int channel)
{
 XWriteFM (voice[0],0x20+InstAd[channel]);
 XWriteFM (voice[1],0x23+InstAd[channel]);
 XWriteFM (voice[2],0x40+InstAd[channel]);
 XWriteFM (voice[3],0x43+InstAd[channel]);
 XWriteFM (voice[4],0x60+InstAd[channel]);
 XWriteFM (voice[5],0x63+InstAd[channel]);
 XWriteFM (voice[6],0x80+InstAd[channel]);
 XWriteFM (voice[7],0x83+InstAd[channel]);
 XWriteFM (voice[8],0xE0+InstAd[channel]);
 XWriteFM (voice[9],0xE3+InstAd[channel]);
 XWriteFM (voice[10],0xC0+channel);
}

void SetUpVoices()
{
fprintf(f,"AdlibSetupData:\n");
 Adlib_SetVoice (Sound_Melodic,0);
 Adlib_SetVoice (Sound_Melodic,1);
 Adlib_SetVoice (Sound_Wave,2);
 Adlib_SetVoice (Sound_Noise,3);
fprintf(f,"db 0,0\n");
}

static int FST[8] = { 0,7,4,3,2,1,1,1 };

void main()
{
int i,n;
f = fopen("SoundDat.inc","w");
SetUpVoices();
SetVolume(0,15);
fprintf(f,"FreqData:\n");
for (i = 0;i < 2048;i++)
	SetFrequency(0,i);
fprintf(f,"InstAddr: db %d,%d,%d,%d\n",InstAd[0],InstAd[1],InstAd[2],InstAd[3]);
fprintf(f,"VolData:\n");
for (i = 0;i < 16;i++)
	fprintf(f," db %d\n",Vol[i] > 0 ? Vol[i]+3 : 0);
fprintf(f,"FSTab:\n");
for (i = 0;i < 256;i++)
	{
	n = (i & 7)*FST[(i & 0x30) >> 4]*((i & 0x08) ? -1 : 1);
	fprintf(f," db %d ; %d\n",n,i);
	}
fclose(f);
for (i = 0;i < 4;i++) SetVolume(i,0);
}

