/************************************************************************/
/*																		*/
/*						All purpose header file							*/
/*																		*/
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

typedef int BOOL;						/* Boolean type */
#define FALSE 		(0)
#define TRUE		(!(FALSE))

#define ASSERT(x)   if (!(x)) Fatal(__LINE__,__FILE__)

#define X86_CARRY	(0x01)				/* 8086 flag masks */
#define X86_HCARRY	(0x10)
#define X86_ZERO	(0x40)

#define CB_EXT		(0x100)				/* Added to opcodes for CB */

#define REG_B		(0)					/* Z80 registers */
#define REG_C		(1)
#define REG_D		(2)
#define REG_E		(3)
#define REG_H		(4)
#define REG_L		(5)
#define REG_HL		(6)
#define REG_A		(7)

void AtomRead(char *Reg86);
void AtomTranslate(void);
void AtomWrite(char *Reg86);
void Fetch(char *Reg86);
void FetchWord(void);
void AddCycles(int Count);
void Data(char *Name,int Size);
void MemRead(char *Reg86);
void MemWrite(char *Reg86);
void ClearFlag(int Mask);
void SetFlag(int Mask);

int	ClaimLabel(int Required);
void Define(int Opcode,int Cycles,char *Command);
void EndDefine(void);
void Fatal(int Line,char *File);
void GetReg(int Number,char *Reg86);
void PutReg(int Number,char *Reg86);
char *GetRegName(int Number);
char *GetZRegName(char *Name);
void Label(char *Lbl);
char *OpLabel(int Opcode);
char *SysLabel(int Label);
void SetNFlag(BOOL Set);
int InString(char *s1,char *s2);
void StarBar(char *s1);
void GenerateVectorTables(void);
void LookJump(char *Table);

char *GetOpName(int Number);
void Operation(int Op);
void LoadFlags(void);
void SaveFlags(void);

void Group1(void);
void Group2(void);
void Group3(void);
void Group4(void);
void Group5(void);
void Group6(void);
void Group7(void);
void Group8(void);
void Group9(void);
void GroupCB(void);

#define Proc(x)		printf("\n%s proc\n",x)
#define EndProc(x)	printf("%s endp\n\n",x)

#define O_ADD		(0)					/* Operations (see document) */
#define O_ADC		(1)
#define O_SUB		(2)
#define O_SBC		(3)
#define O_AND		(4)
#define O_XOR		(5)
#define O_OR 		(6)
#define O_CP 		(7)
#define O_RLC		(8)
#define O_RRC		(9)
#define O_RL 		(10)
#define O_RR 		(11)
#define O_SLA		(12)
#define O_SRA		(13)
#define O_SWAP		(14)
#define O_SRL		(15)
#define O_SETI		(16)
#define O_SETR		(17)
#define O_RESI		(18)
#define O_RESR		(19)
#define O_BIT 		(20)
#define O_INC8		(22)
#define O_DEC8		(23)
#define O_PUSHCX	(24)
#define O_POPCX		(25)
#define O_ADD16		(26)
#define O_UMF		(27)
#define O_LMF		(28)
#define O_SETZ		(29)
#define O_SEXA		(30)

#define O_LASTOP	(O_SEXA)

#define IS8086REG(x) (InString("al\0bl\0bh\0cl\0ch\0dl\0dh\0",x) != -1)
#define ISZ80REG(x) (InString("a\0b\0c\0d\0e\0h\0l\0f\0af\0bc\0de\0hl\0pc\0sp\0Cycles\0Iff\0NFlag",x) != -1)

#define ISSTRING(s) ((s) != NULL && *(s) != '\0')
#define IS8BITREG(n) ((n) >= 0 && (n) < 8)
#define ISBOOL(b)	((b) == TRUE || (b) == FALSE)