/************************************************************************/
/*                                                                      */
/*						Instruction Generation							*/
/*                                                                      */
/************************************************************************

Generates : Groups 5,6,7

*/

#include "cpu.h"

static void PostIncDec(int Reg);
static void DoGroup6Code(int Op);
static void DoGroup6Fetch(int Op);

/************************************************************************/
/*                                                                      */
/*						Group 5 - 16 bit operations						*/
/*                                                                      */
/************************************************************************/

static char *Reg16a[] = { "bc","de","hl","sp" };
static char *Reg16b[] = { "(bc)","(de)","(hl)+","(hl)-" };
static char *Reg16c[] = { "bc","de","hl","hl" };

void Group5(void)
{
int Reg;
char Descr[16];
for (Reg = 0;Reg < 4;Reg++)				/* 4 register pairs */
	{
	/**** Do ld reg16,#const ****/
	sprintf(Descr,"ld %s,#nnnn",Reg16a[Reg]);
	Define(0x1+Reg*16,12,Descr);
	FetchWord();                		/* Get the value & store it */
	printf("    mov [word ptr %s],bx\n",GetZRegName(Reg16a[Reg]));
	EndDefine();

	/**** Do inc reg16 ****/
	sprintf(Descr,"inc %s",Reg16a[Reg]);
	Define(0x3+Reg*16,8,Descr);
	printf("    inc [word ptr %s]\n",GetZRegName(Reg16a[Reg]));
	EndDefine();

	/**** Do dec reg16 ****/
	sprintf(Descr,"dec %s",Reg16a[Reg]);
	Define(0xB+Reg*16,8,Descr);
	printf("    dec [word ptr %s]\n",GetZRegName(Reg16a[Reg]));
	EndDefine();

	/**** Do add hl,reg16 ****/
	sprintf(Descr,"add hl,%s",Reg16a[Reg]);
	Define(0x9+Reg*16,8,Descr);
	printf("    mov bx,[word ptr %s]\n",GetZRegName(Reg16a[Reg]));
	Operation(O_ADD16);
	EndDefine();

	/**** Do ld a,(reg16) ****/
	sprintf(Descr,"ld a,%s",			/* Note these are (BC) (DE) (HL+) */
						Reg16b[Reg]);	/* and (HL-) */
	Define(0xA+Reg*16,8,Descr);
	printf("    mov bx,[word ptr %s]\n",/* Get address to read */
							GetZRegName(Reg16c[Reg]));
	MemRead("al");						/* Do the read */
	PutReg(REG_A,"al");
	PostIncDec(Reg);					/* Do the HL+ or HL- if required */
	EndDefine();

	/**** Do ld (reg16),a ****/
	sprintf(Descr,"ld %s,a",			/* These are (HL+) and (HL-) too */
						Reg16b[Reg]);	/* same as above but backwards */
	Define(0x2+Reg*16,8,Descr);
	printf("    mov bx,[word ptr %s]\n",GetZRegName(Reg16c[Reg]));
	GetReg(REG_A,"al");
	MemWrite("al");
	PostIncDec(Reg);
	EndDefine();
	}
}

/************************************************************************/
/*                                                                      */
/*			Generate post-increment/post-decrement code					*/
/*                                                                      */
/************************************************************************/

static void PostIncDec(int Reg)
{
ASSERT(Reg >= 0 && Reg < 4);
if (Reg == 2) printf("    inc [word ptr %s]\n",GetZRegName("HL"));
if (Reg == 3) printf("    dec [word ptr %s]\n",GetZRegName("HL"));
}

/************************************************************************/
/*                                                                      */
/*				Group 6 - control transfer operations					*/
/*                                                                      */
/************************************************************************/

static char *XFName[] = { "jr","jp","call","ret" };	/* Name of instruction */
static int XFBase[] = 	{ 0x20,0xC2,0xC4,0xC0 };	/* Conditionals base */
static int XFUncond[] = { 0x18,0xC3,0xCD,0xC9 };	/* Unconditional equiv */
static int XFDone[]   = { 12,16,24,16 };			/* Cycles when done */
static int XFFail[]   = { 8,12,12,8 };				/* Cycles when not done */
static char *XFCond[] = { "nz","z","nc","c" };		/* condition name */

void Group6(void)
{
int Cond,Op,Lbl;
char Descr[16];
for (Op = 0;Op < 4;Op++)				/* for jr,jp,call and ret */
	{
	sprintf(Descr,"%s %s",				/* Unconditional one first */
			XFName[Op],(Op < 3) ? "nnnn" : "");
	Define(XFUncond[Op],XFDone[Op],Descr);
	DoGroup6Fetch(Op);					/* Do the EAC */
	DoGroup6Code(Op);					/* and the instruction */
	EndDefine();

	for (Cond = 0;Cond < 4;Cond++)		/* for each of the conditions */
		{
		Lbl = ClaimLabel(1);
		sprintf(Descr,"%s %s%s",
					XFName[Op],XFCond[Cond],(Op == 3) ? "":",nnnn");
		Define(XFBase[Op]+Cond*8,XFFail[Op],Descr);
		DoGroup6Fetch(Op);				/* Do the EAC */

		printf("    test ah,%d\n",		/* Check the appropriate bit in F */
						(Cond < 2) ? X86_ZERO : X86_CARRY);
		printf("    j%s short %s\n",	/* skip if failing */
						(Cond % 2 == 0) ? "nz" : "z",SysLabel(Lbl));

		AddCycles(XFDone[Op]-XFFail[Op]); /* Added cycles cos we passed */
		DoGroup6Code(Op);				/* Do the instruction */
		Label(SysLabel(Lbl));
		EndDefine();
		}
	}
}

/************************************************************************/
/*                                                                      */
/*					Do the appropriate code for Group 6					*/
/*                                                                      */
/************************************************************************/

static void DoGroup6Code(int Op)
{
switch(Op)
	{
	case 0: printf("    mov bx,[%s]\n",	/* Jump Relative */
								GetZRegName("PC"));
			Operation(O_SEXA);			/* Do a relative ADD AL to PC */
			printf("    mov [%s],bx\n",
							GetZRegName("PC"));
			break;

	case 1:	printf("    mov [%s],bx\n",	/* Jump absolute */
								GetZRegName("PC"));
			break;

	case 2:	printf("    push bx\n");	/* Call - save new address */
			printf("    mov cx,[%s]\n",	/* Get PC & Push it */
							GetZRegName("PC"));
			Operation(O_PUSHCX);
			printf("    pop bx\n");		/* Restore old address & jump */
			printf("    mov [%s],bx\n",GetZRegName("PC"));
			break;

	case 3:	Operation(O_POPCX);			/* Return */
			printf("    mov [%s],cx\n",GetZRegName("PC"));
			break;

	default:ASSERT(FALSE);
			break;
	}
}

/************************************************************************/
/*                                                                      */
/*					Do the fetches for an operation						*/
/*                                                                      */
/************************************************************************/

static void DoGroup6Fetch(int Op)
{
if (Op == 0) Fetch("al");				/* jr fetches byte to al */
if (Op == 1 || Op == 2) FetchWord();	/* jp & call fetch word to bx */
										/* ret has no EAC */
}

/************************************************************************/
/*                                                                      */
/*				Generate Group 7 Instructions (Stack)					*/
/*                                                                      */
/************************************************************************/

void Group7(void)
{
int r;
char *PPReg[] = { "bc","de","hl","af" };/* Registers used */
char Descr[16];
for (r = 0;r < 4;r++)					/* for each register pair */
	{
	/* PUSH xx */
	sprintf(Descr,"push %s",PPReg[r]);
	Define(0xC5+r*16,16,Descr);
	if (r == 3) Operation(O_UMF);		/* if pushing AF make memory F correct */
	printf("    mov cx,[word ptr %s]\n",GetZRegName(PPReg[r]));
	Operation(O_PUSHCX);
	EndDefine();

	/* POP xx */
	sprintf(Descr,"pop %s",PPReg[r]);
	Define(0xC1+r*16,12,Descr);
	Operation(O_POPCX);
	printf("    mov [word ptr %s],cx\n",GetZRegName(PPReg[r]));
	if (r == 3) Operation(O_LMF);		/* if popping AF load F into registers */
	EndDefine();
	}
}

/************************************************************************/
/*                                                                      */
/*							Group 8 - Restarts							*/
/*                                                                      */
/************************************************************************/

void Group8(void)
{
int r;
char Desc[32];
for (r = 0;r < 8;r++)					/* 8 Restarts 00..38h */
	{
	sprintf(Desc,"rst %xh",r*8);
	Define(0xC7+r*8,16,Desc);
	printf("    mov cx,[%s]\n",			/* Push PC onto the stack */
					GetZRegName("PC"));
	Operation(O_PUSHCX);
	printf("    mov bx,%d\n",r*8);		/* Do the jump */
	printf("    mov [%s],bx\n",GetZRegName("PC"));
	EndDefine();
	}
}