; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *                     Main Gameboy emulator program                   *
; *                                                                     *
; ***********************************************************************
; *                                                                     *
; *             Written by Paul Robson 1997. All rights reserved.       *
; *                                                                     *
; *                         19 May 1997 - date                          *
; *                                                                     *
; * This program and its source codes are Copyright Paul Robson (C)     *
; * 1997 except where otherwise stated. License is granted to freely    *
; * copy and distribute the source & object codes, provided this is     *
; * done without charge. Permission is granted to modify the source     *
; * code provided the original author is acknowledged.                  *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************
;
;  Build 1.047   : First program & Jens' Drawing Demo both work correctly.
;        1.106   : Vertical scrolling bug fixed.
;        1.123   : Horizontal scrolling works.
;        1.190   : Interrupts tested on GBINTS.
;        1.192   : Keyboard handler installed.
;        1.212   : Window going
;        1.235   : Frame skip correctly operating
;        1.369   : Added serial interrupt,lots of testing & optimisation
;        1.413   : Fixed Tetris/DrMario bug (incrementing Divider reg)
;        1.452   : 32k ROMs all work ok, except there's no sprites.
;                  Final timing bugs (hopefully) fixed.
;        1.526   : Sprites now added & working (8+16 X+Y Flip !)
;        1.577   : Support for MBC1+2 ROMs (Super Mario Bros 1 working,so
;                  does Xenon2 Bubble Bobble Jr Pacman I've only tried these.
;                  Zelda,ToyStory Return of the Jedi & Donkey Kong 3 all work ok.
;        1.593   : Added support for 4 RAM Banks. Zoomar now works properly.
;                  Everything tried worked ok.
;        1.708   : First version with sound. Problems with the noise channel
;                  No envelopes.
;        1.772   : Loads files from CLI,Sync added + odd timing stuff...
;        1.786   : Fixed the Tetris Flash bug
;        1.808   : Down to 8 failures out of 160 (not including window bug)
;        1.833   : Battery Backed up added,compatible with VGB.
;        1.844   : Added benchmarking file
;        1.862   : Can load configuration (.G97) Files
;        1.897   : Added Int60 & made it disableable. Lawnmower man and
;                  Double Dragon 3 now work,as should any games that lock
;                  at the start.
;        1.921   : Proper interrupt handler. Games failing now 6 out of 188.
;        1.934   : Final Fantasy Legend 2 works. (Int raise changed to OR)
;        1.985   : Fixed Bubble Bobble. (then unfixed it)
;======= 1.989   : First External Release (proper)
;        2.004   : VBlank Interrupt moved to correct position. Daedopus now
;                  works.
;        2.010   : Added Recursive Interrupt block for Bubble Bobble 1
;        2.014   : Correctly mapped C000-FDFF twice as per Dr Pan's document
;        2.021   : Autofire added.
;        2.091   : Fixed INT9 bug - now runs correctly using WDOSX built in
;                  DOS Extender.
;        2.145   : Added sweeps to sound.
;        2.158   : Added support for 32k NVRAM
;        2.189   : Added Autofire Control & Text Display routines.
;        2.222   : Added Game Genie Cheat Codes
;        2.226   : Keyboard byte $FF00 is not updated on the fly now
;                  only when $FF00 is written to.
;        2.247   : Added Joystick support.
;        2.271   : Added the 'Gameboy Alike' border.
;        2.362   : Added prioritisable sprites.
;        2.425   : Added support for 4 button joysticks.
;        2.446   : Colour Schemes supported (via translator)
;        2.542   : Fixed Donkey Kong bug. (looks odd but playable)
;        2.620   : Reset button added
;        2.641   : Added Pause Button / Display Mode switch. Changed
;                  default drawing order of sprites for Zelda IV
;        2.661   : In game save & Load working.
;        2.699   : Ability to Load 256-colour BMP files as background
;        2.731   : Added painting.
;        2.738   : Fix for 'Frank Thornton Baseball' (VBLANK Int when scr off)
;        2.752   : Cart RAM saved in F7 Save
;                  Cart RAM moved into Memory space using first 32k
;                  Current RAM Page set correctly and fix for old types.
;                  Reset reloads BB RAM
;        2.764     Added Or Interrupts (fixed Waverace)
;        2.770     Added New sound code
;        2.795     Fixed Window alt tile bug (used bit 4 of LCDCONT)
;        2.816     Fixed Step Over key
;        2.837     Fix for Gauntlet (shifting the serial data left)
;        2.843     Pacman Sound bug fixed
;        2.851     Joystick Uses BIOS rather than direct access
;        2.875     Hudson Hawk Fix. More readable font.
;        2.882     Early VBI for Addams Familty
;        2.885     Fixed Alleyway (serial in floats high)
;        2.890     Fixed R-Type 2.
;
;                  Kid Niki (wierd,requests odd LCD status interrupt)
;                  Snow Bros (disappearing enemies go off side of screen)
;                  Final Fantasy Adventure (menu screens bastard)
;                  Darkwing Duck (can't get off the map - ignores A/B)
;                  Astro Rabby (game resets)
;                  Foreman Boxing (corrupt display)
;                  Momotaro Thunderbolt 2 (exits)
;                  GUI
;                  Sprite Sorting
;
.386p 
.model flat
.code 
include Gb.Inc

Start:
        mov     [FileName],offset Empty ; Default File name is an empty string
        cmp     esi,2                   ; check ARGC
        jnz     StartProgram
        add     edi,4                   ; get the filename
        mov     eax,[edi]               ; pointer
        mov     [FileName],eax          ; & save it for later use.
StartProgram:
        mov     ah,0                    ; switch to 320x200 VGA mode
        mov     al,013h
        int     010h
        call    Initialise              ; set everything up
        call    UpdateKeyboardByte      ; scan the keyboard
        call    UpdateFF00              ; set keyboard register appropriately
        call    ReadPCTimer             ; reset the benchmark counter
        mov     [BMStart],ax
Main1:  call    DisplayFrame            ; Do one tick = several frames
        call    AccessGUI
        call    ToggleDisplayMode
        call    CheckLoadSave
        call    CheckPaint
        mov     eax,[ResetKey]
        mov     bl,[byte ptr KeyTable+eax]
        test    bl,128
        jnz     ResetEmulator
        call    IsEscape                ; if escape not pressed
        cmp     al,0
        jz      Main1                   ; loop back
        call    ReadPCTimer
        mov     [BMEnd],ax              ; save the end timer value

ifdef PSRBuild
        mov     edx,offset BMFileName   ; write out the benchmark file
        mov     eax,offset BMStart      ; with start/frames/end
        mov     ecx,6
        call    SaveFile
endif
        jmp     Finish

GoDOS:  call    Registers
Finish: call    Terminate               ; otherwise exit
        mov     ah,0                    ; switch to text VGA mode
        mov     al,3h
        int     010h

        mov     edx,offset TermText     ; print the byebye message
        mov     ah,09
        int     21h
        mov     ax,04c00h               ; return to MSDOS
        int     21h

ResetEmulator:
        mov     bl,[byte ptr KeyTable+eax]
        test    bl,128
        jnz     ResetEmulator

        call    Reset                   ; reset the CPU
        call    ResetData               ; reset the BSS data
        call    ResetMapping            ; reset the mapping
        call    ResetMappingData        ; and the masks
        call    ClearMemory             ; clear memory
        mov     eax,Ext_SAV             ; load sav file
        call    FileNameExt
        lea     esi,CartridgeRAM
        call    LoadFile                ; ignore errors.
        jmp     Main1

        call    Terminate
        jmp     StartProgram

TermText:
        db      10
        db      'GB97 Gameboy Emulator in 32-bit Assembler',13,10,10
        db      'This emulator and its source are freeware. However, if you wish to develop',13,10
        db      'the emulator further please ask permission, and keep me informed of updates.',13,10,10
        db      'Code by      : Paul Robson (c) 1997 (autismuk@aol.com)',13,10
        db      'Assembler    : Borland Turbo Assembler 4.0',13,10
        db      'C Compiler   : Borland Turbo C 2.0',13,10
        db      'DOS Extender : WDosX 0.93 by Michael Tippich',13,10,10
        include BuildDat.Inc
        db      13,10,'$'

Empty:  db      'dkong',0             ; empty file name

BMFileName:db   'BenchMrk.Dat',0        ; name of the benchmark file

end start

