/************************************************************************/
/*                                                                      */
/*		Gameboy emulated CPU Generator v1.0 (C) Paul Robson 1997		*/
/*                                                                      */
/************************************************************************

v1.0 (19-May-97)		First version complete & tested.
						Runs at about 14Mhz on my machine.

	 (21-May-97)		RCL/RCR/ROL/ROR wrong way round.
						BIT setting Z flag incorrectly.

	 (22-May-97)		Interrupt clears IFLAGS bit

	 (23-May-97)		Fixed interrupt calling bug
						Fixed timer rate.

	 (26-May-97)		Carry cleared on AND/OR/XOR
						Inc & Dec preserve Carry.
						Scf,Ccf set H and N correctly.
						Swap doesn't set Z flag.
						Timer rate change was resetting the counter.
						Now only resets counter if count is greater
						than the count rate.

*/

#include "cpu.h"

static void Registers(void);
static void Utilities(void);
static void AddCode(void);
static void DumpMemory(int n);
static void ResetCode(void);

void main()
{
#ifndef TESTING
StarBar(" Z80 (Gameboy) Emulation. Automatic Code,don't modify it");
printf("\n\n\n");
Registers();
ResetCode();
Group1();
Group2();
Group3();
Group4();
Group5();
Group6();
Group7();
Group8();
Group9();
GroupCB();

GenerateVectorTables();
MappingTable();
#endif
Utilities();
}

/************************************************************************/
/*                                                                      */
/*						Create the registers							*/
/*                                                                      */
/************************************************************************/

static void Registers()
{
static char *SplitReg[] = { "af","bc","de","hl" };
int i,n;
char Temp[2];


StarBar("Register Definitions");
printf(".data \n");
printf("CPURStart:\n");
for (i = 0;i < 4;i++)
	{
	Label(GetZRegName(SplitReg[i]));
	for (n = 1;n >= 0;n--)
		{
		Temp[0] = SplitReg[i][n];Temp[1] = '\0';
		Data(GetZRegName(Temp),1);
		}
	}

Data(GetZRegName("PC"),2);
Data(GetZRegName("SP"),2);
Data(GetZRegName("Cycles"),2);

Data(GetZRegName("IFF"),1);
Data(GetZRegName("NFLAG"),1);

printf("TimerCount dw 1\n");			/* Timer Counter */
printf("TimerRate dw 16\n");			/* Timer Rate */
printf("CycleCount dw 0\n");			/* No of cycles till end of this go */
printf("TraceFlag db 0\n");				/* Trace flag (if in use) */
printf("OrgIFF    db 0\n");				/* IFF at interrupt time */
printf("Last dw 0\n");

printf("CPUREnd:\n");

printf(".code\n\n");
}

static void Utilities(void)
{
int n,c;
char *RegDump[] = { "PC","A","F","B","C","D","E","H","L","SP","IFF",NULL };
char *p;
AddCode();

/* Generate an interrupt. Checks for EI and FF0F != 0 should have been
   done. Also deals with waking up CPU */

StarBar("Generate an interrupt. EI and FF0F tested");
Proc("Interrupt");
printf(" push eax ebx ecx edx esi\n");  		/* save everything */

printf(" mov al,[%s]\n",GetZRegName("IFF"));
printf(" mov [OrgIFF],al\n");

n = ClaimLabel(1);
printf(" test [%s],080h\n",GetZRegName("IFF"));	/* if processor in wait mode */
printf(" jz %s\n",SysLabel(n));
printf(" inc [%s]\n",GetZRegName("PC"));		/* bump the PC */
Label(SysLabel(n));
printf(" and [%s],07Eh\n",GetZRegName("IFF"));	/* clear wait & ie bits */
printf(" mov  cx,[%s]\n",GetZRegName("PC"));	/* save the PC */
Operation(O_PUSHCX);
printf(" mov cl,[Byte Ptr Memory+0FF0Fh]\n");	/* get the IFLAGS */
for (c = 0;c < 5;c++)							/* check whodunnit */
	{											/* increasing priority */
	n = ClaimLabel(1);
	printf(" test cl,%d\n",(1 << c));
	printf(" jz %s\n",SysLabel(n));
	printf(" mov ax,0%04xh\n",0x0040+c*8);      /* put addr in AX */
	Label(SysLabel(n));
	}
for (c = 0;c < 5;c++)							/* figure out what bit to */
	{											/* reset */
	n = ClaimLabel(1);
	printf(" cmp ax,0%04xh\n",0x0040+c*8);
	printf(" jne %s\n",SysLabel(n));
	printf(" and [Byte Ptr Memory + 0FF0Fh],0%02xh\n",255-(1 << c));
	Label(SysLabel(n));
	}

printf(" mov [%s],ax\n",GetZRegName("PC"));		/* and go there */
printf(" pop esi edx ecx ebx eax\n");
printf(" ret\n");
EndProc("Interrupt");

/* Execute AX CPU Cycles */

StarBar("Execute AX CPU Cycles");
Proc("ExecuteCycles");
printf("    mov [CycleCount],ax\n");	/* Set cycle count */
Operation(O_LMF);						/* Load Memory Flags into Emulation */
n = ClaimLabel(6);


printf("    mov edi,offset %s\n",		/* EDI = Return Address for Jump */
							SysLabel(n+5));
printf("    call DebugPatchIn\n");		/* Debugger patch */

Label(SysLabel(n));         			/* Main loop */

LookJump("OPV00");						/* Call the executor */
Label(SysLabel(n+5));					/* We return to here */

#ifdef TRACING							/* Debugging code */
printf("    cmp [TraceFlag],0\n");
printf("    jz %s\n",SysLabel(n+3));
printf("    call Registers\n");
Label(SysLabel(n+3));
#endif

/* Check for interrupts */
printf("    mov al,[byte ptr Memory+0FF0Fh]\n");
printf("    and al,255\n");
printf("    jnz  %s\n",SysLabel(n+2));	/* Interrupt occurred,check it out */
Label(SysLabel(n+4));

printf("    mov  bx,[%s]\n",			/* Get cycles */
					GetZRegName("Cycles"));
printf("    sub [TimerCount],bx\n");	/* Timer elapsed ? */
printf("    js %s\n",SysLabel(n+1));	/* yes, do that bit */
printf("    sub [CycleCount],bx\n");	/* done enough cycles.... */
printf("    jnc %s\n",SysLabel(n));		/* no,do another instruction */
Operation(O_UMF);   					/* Save CPU flags to memory */
printf("    ret\n");

Label(SysLabel(n+1));					/* Timer has timed out */
printf("    mov cx,[TimerRate]\n");		/* Reset the timer counter */
printf("    mov [TimerCount],cx\n");
printf("    call TimedOut\n");			/* Call the timer check routine */
printf("    sub [CycleCount],bx\n");	/* done enough cycles.... */
printf("    jnc %s\n",SysLabel(n));		/* no,do another instruction */
Operation(O_UMF);   					/* Save CPU flags to memory */
printf("    ret\n");

Label(SysLabel(n+2));					/* Come here when FF0F non-zero */
printf("    test [%s],081h\n",			/* if interrupts disabled ? */
					GetZRegName("IFF"));
printf("    jz  %s\n",SysLabel(n+4));	/* if so,resume execution */
printf("    call Interrupt\n");			/* Generate the appropriate int */
printf("    jmp %s\n",SysLabel(n+4));	/* and resume */

EndProc("ExecuteCycles");

Proc("UpdateCPUFlagByte");
Operation(O_UMF);
printf("    ret\n");
EndProc("UpdateCPUFlagByte");

/* reset the timer rate. if the count is greater than the rate, the
   count is reset to the rate... */

StarBar("Set the timer rate to AX Cycles/Call");
n = ClaimLabel(1);
Proc("SetTimerRate");
printf("    mov [TimerRate],ax\n");		/* save the timer rate */
printf("    cmp [TimerCount],ax\n");	/* if count greater than rate */
printf("    jb short %s\n",SysLabel(n));/* and rate slowing could have a */
printf("    mov [TimerCount],ax\n");	/* long wait so reset counter */
Label(SysLabel(n));
printf("    ret\n");
EndProc("SetTimerRate");

Proc("AtomTranslate");					/* Provides ATOMTRANSLATE to the */
AtomTranslate();						/* emulation core, primarily for */
printf("    ret\n");					/* DMA Write */
EndProc("AtomTranslate");

StarBar("Display the registers on stdout");
Proc("Registers");
printf(" push eax ebx ecx edx esi edi ebp\n");
Operation(O_UMF);
printf(" mov al,%d \n call DebugChar\n",'Z');
printf(" mov al,%d \n call DebugChar\n",'8');
printf(" mov al,%d \n call DebugChar\n",'0');
printf(" mov al,%d \n call DebugChar\n",'>');
n = 0;
while (RegDump[n] != NULL)
	{
	p = RegDump[n];
	printf(" mov al,%d\n call DebugChar\n",*p);
	if (*(p+1) != '\0')
		printf(" mov al,%d\n call DebugChar\n",*(p+1));
	printf(" mov al,58\n call DebugChar\n",*p);
	if (strlen(p) != 2)
		printf(" mov al,[%s]\n call DebugByte\n",GetZRegName(p));
	else
		printf(" mov ax,[%s]\n call DebugWord\n",GetZRegName(p));
	printf(" mov al,32\n call DebugChar\n",*p);

	n++;
	}
for (n = 0;n < 4;n++)
	{
	c = ClaimLabel(1);
	printf(" mov al,%d\n",*("ZNHC"+n));
	printf(" test [%s],%d\n",GetZRegName("F"),0x80 >> n);
	printf(" jnz short %s\n",SysLabel(c));
	printf(" mov al,'_'\n");
	Label(SysLabel(c));
	printf(" Call DebugChar\n");
	}
printf(" mov bx,[%s]\n",GetZRegName("PC"));
AtomTranslate();
for (n = 0;n < 3;n++)
	{
	printf(" mov al,32\n call DebugChar\n",*p);
	AtomRead("al");
	printf(" call DebugByte\n");
	printf(" inc esi\n");
	}
printf(" mov al,10\n call DebugChar\n");
printf(" mov al,13\n call DebugChar\n");
/*
DumpMemory(0xFF00);
DumpMemory(0xFF40);
DumpMemory(0xFFF0);
*/
printf(" pop ebp edi esi edx ecx ebx eax\n");
printf(" ret\n");
EndProc("Registers");

}

static void DumpMemory(int n)
{
int i;
printf(" mov ax,%d\n call DebugWord\n",n);
for (i = 0;i < 16;i++)
	{
	printf(" mov al,32\n call DebugChar\n");
	printf(" mov bx,%d\n",n+i);
	MemRead("al");
	printf(" call DebugByte\n");
	}

printf(" mov al,10\n call DebugChar\n");
printf(" mov al,13\n call DebugChar\n");
}


static void AddCode(void)
{
static char *Code[] = {
"DebugChar Proc",
" push eax",
" push edx",
" mov  dl,al",
" mov  ah,2",
" int  21h",
" pop  edx",
" pop  eax",
" ret",
"EndP DebugChar",

"DebugNibl Proc",
" push eax",
" and eax,15",
" cmp eax,10",
" jb DebugNibl1",
" add al,7",
"DebugNibl1:",
" add al,48",
" call DebugChar",
" pop eax",
" ret",
"EndP DebugNibl",

"DebugByte Proc",
" push eax",
" shr al,4",
" call DebugNibl",
" pop eax",
" call DebugNibl",
" ret",
"EndP DebugByte",

"DebugWord Proc",
" push eax",
" mov al,ah",
" call DebugByte",
" pop eax",
" call DebugByte",
" ret",
"EndP DebugWord",

NULL };

int i = 0;

while (Code[i] != NULL) printf("%s\n",Code[i++]);
}

static void ResetCode(void)
{
StarBar("CPU Reset");
Proc("CPUReset");
printf(" mov [%s],100h\n",GetZRegName("PC"));
printf(" mov [%s],0FFFEh\n",GetZRegName("SP"));
printf(" mov [word ptr %s],001B0h\n",GetZRegName("AF"));
printf(" mov [word ptr %s],00013h\n",GetZRegName("BC"));
printf(" mov [word ptr %s],0014Dh\n",GetZRegName("HL"));
printf(" mov [word ptr %s],000D8h\n",GetZRegName("DE"));
printf(" ret\n");
EndProc("CPUReset");
}
