/************************************************************************/
/*                                                                      */
/*							Operation generation						*/
/*                                                                      */
/************************************************************************/

#include "cpu.h"

static void UpdateMFR(void);
static void LoadMFR(void);

/************************************************************************/
/*                                                                      */
/*						Move flags to and from AH						*/
/*                                                                      */
/************************************************************************/

void LoadFlags(void)
{
printf("    sahf\n");
}

void SaveFlags(void)
{
printf("    lahf\n");
}

/************************************************************************/
/*                                                                      */
/*						Return name of an operation						*/
/*                                                                      */
/************************************************************************/

static char *OpNames[] = {

	"add","adc","sub","sbc","and","xor","or","cp",
	"rlc","rrc","rl","rr","sla","sra","swap","srl",
	"seti","setr","resi","resr","bit","---","inc8","dec8",
	"pushcx","popcx","add16","umf","lmf","setz","sexa"
	};

char *GetOpName(int Number)
{
ASSERT(Number >= 0 && Number <= O_LASTOP);
return(OpNames[Number]);
}

/************************************************************************/
/*                                                                      */
/*						Generate operation code							*/
/*                                                                      */
/************************************************************************/

void Operation(int Op)
{
char *Cmd;
int RLabel,Lbl;
ASSERT(Op >= 0 && Op <= O_LASTOP);
switch(Op)
	{
	case O_ADD:        					/* Basic arithmetic operations */
	case O_ADC:
	case O_SUB:
	case O_SBC:
	case O_CP:	Cmd = GetOpName(Op);	/* Get the 8086 equivalent */
				if (Op == 3) Cmd = "sbb";
				if (Op == 7) Cmd = "cmp";
				LoadFlags();
				printf("    %s [%s],al\n",Cmd,GetZRegName(GetRegName(REG_A)));
				SaveFlags();
				SetNFlag(Op == 2 || Op == 3);
				break;
	case O_AND: 						/* Logic operations */
	case O_OR:
	case O_XOR:	Cmd = GetOpName(Op);	/* Get the 8086 equivalent */
				if (Op == 3) Cmd = "sbb";
				printf("    %s al,[%s]\n",Cmd,GetZRegName(GetRegName(REG_A)));
				printf("    mov [%s],al\n",GetZRegName(GetRegName(REG_A)));
				Operation(O_SETZ);
				SetNFlag(0);
				ClearFlag(X86_CARRY);
				break;

	case O_RL:
	case O_RLC:
	case O_RR:
	case O_RRC:
	case O_SLA:
	case O_SRA:
	case O_SRL:	Cmd = GetOpName(Op);
				if (Op == O_SLA) Cmd = "sal";
				if (Op == O_SRL) Cmd = "shr";
				if (Op == O_SRA) Cmd = "sar";
				if (Op == O_RLC) Cmd = "rol";
				if (Op == O_RRC) Cmd = "ror";
				if (Op == O_RL) Cmd = "rcl";
				if (Op == O_RR) Cmd = "rcr";
				LoadFlags();
				printf("    push ax\n");
				printf("    %s al,1\n",Cmd);
				SaveFlags();
				printf("    pop cx\n");
				printf("    and ah,%d\n",X86_CARRY);
				printf("    and ch,%d\n",255-X86_CARRY);
				printf("    or ah,ch\n");
				SetNFlag(0);
				ClearFlag(X86_HCARRY);
				break;

	case O_SWAP:Cmd = GetOpName(Op);
				printf("    ror al,4\n",Cmd);
				SetNFlag(0);
				break;

	case O_INC8:						/* 8 bit increment + decrement */
	case O_DEC8:
				LoadFlags();
				printf("    push ax\n");
				printf("    %s al\n",(Op == O_INC8 ? "inc":"dec"));
				SaveFlags();
				SetNFlag(Op == O_DEC8);
				printf("    pop cx\n");
				printf("    and ch,%d\n",X86_CARRY);
				printf("    and ah,%d\n",255-X86_CARRY);
				printf("    or ah,ch\n");
				break;

	case O_ADD16:						/* 16 bit add of bx to HL */
				ClearFlag(X86_CARRY+X86_HCARRY);
				printf("    mov cx,bx\n");	/* start hc calculation */
				printf("    xor cx,[word ptr %s]\n",GetZRegName("HL"));

				printf("    add [word ptr %s],bx\n",GetZRegName("HL"));
				RLabel = ClaimLabel(2);
				printf("    jnc %s\n",SysLabel(RLabel));
				SetFlag(X86_CARRY);
				Label(SysLabel(RLabel));
				printf("    xor cx,[word ptr %s]\n",GetZRegName("HL"));
				printf("    and cx,01000h\n");
				printf("    jz %s\n",SysLabel(RLabel+1));
				SetFlag(X86_HCARRY);
				Label(SysLabel(RLabel+1));
				break;

	case O_SEXA:printf("    push ax\n");/* Sign extend add al to bx */
				printf("    cbw\n");
				printf("    add bx,ax\n");
				printf("    pop ax\n");
				break;

	case O_PUSHCX:						/* Push CX on the stack */
				#ifdef QUICK
					printf("    sub [%s],2\n",GetZRegName("SP"));
					printf("    mov bx,[%s]\n",GetZRegName("SP"));
					AtomTranslate();
					printf("    mov [esi],cx\n");
				#else
					printf("    sub [%s],2\n",GetZRegName("SP"));
					printf("    mov bx,[%s]\n",GetZRegName("SP"));
					MemWrite("cl");
					printf("    inc bx\n");
					MemWrite("ch");
				#endif
				break;

	case O_POPCX:						/* Pop CX off the stack */
				#ifdef QUICK
					printf("    mov bx,[%s]\n",GetZRegName("SP"));
					printf("    add [%s],2\n",GetZRegName("SP"));
					AtomTranslate();
					printf("    mov cx,[esi]\n");
				#else
					printf("    mov bx,[%s]\n",GetZRegName("SP"));
					printf("    add [%s],2\n",GetZRegName("SP"));
					MemRead("cl");
					printf("    inc bx\n");
					MemRead("ch");
				#endif
				break;

	case O_UMF: UpdateMFR();			/* Update the memory flag register */
				break;					/* from AH and the N flag */

	case O_LMF:	LoadMFR();				/* Load AH and N flag register from */
				break;					/* the memory flag register */

	case O_BIT:	Lbl = ClaimLabel(1);	/* set z flag if al & cl is nonzero */
				ClearFlag(X86_ZERO);
				SetNFlag(FALSE);
				SetFlag(X86_HCARRY);
				printf("    and al,cl\n");
				printf("    jnz short %s\n",SysLabel(Lbl));
				SetFlag(X86_ZERO);
				Label(SysLabel(Lbl));
				break;

	case O_SETZ:Lbl = ClaimLabel(1);	/* set z flag if al is zero */
				ClearFlag(X86_ZERO);
				printf("    or al,al\n");
				printf("    jnz short %s\n",SysLabel(Lbl));
				SetFlag(X86_ZERO);
				Label(SysLabel(Lbl));
				break;

	default:	ASSERT(FALSE);
				break;
	}
}

/************************************************************************/
/*                                                                      */
/*				Update memory flag register from NFlag and AH			*/
/*                                                                      */
/************************************************************************/

static int TestBit[] = { X86_CARRY,X86_HCARRY,X86_ZERO };
static int SetBit[] = { 0x10,0x20,0x80 };

static void UpdateMFR(void)
{
static int b,Lbl;

printf("    mov al,[%s]\n",GetZRegName("NFlag"));
printf("    and al,040h\n");
for (b = 0;b < 3;b++)
	{
	Lbl = ClaimLabel(1);
	printf("    test ah,0%02xh\n",TestBit[b]);
	printf("    jz short %s\n",SysLabel(Lbl));
	printf("    or al,0%02xh\n",SetBit[b]);
	Label(SysLabel(Lbl));
	}
printf("    mov [%s],al\n",GetZRegName("F"));
}

static void LoadMFR(void)
{
int b,Lbl;
printf("    xor ah,ah\n");
printf("    mov al,[%s]\n",GetZRegName("F"));
for (b = 0;b < 3;b++)
	{
	Lbl = ClaimLabel(1);
	printf("    test al,0%02xh\n",SetBit[b]);
	printf("    jz short %s\n",SysLabel(Lbl));
	printf("    or ah,0%02xh\n",TestBit[b]);
	Label(SysLabel(Lbl));
	}
Lbl = ClaimLabel(1);
printf("    and al,040h\n");
printf("    jz  short %s\n",SysLabel(Lbl));
printf("    mov al,0FFh\n");
Label(SysLabel(Lbl));
printf("    mov [%s],al\n",GetZRegName("NFlag"));
}