; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *                        Sprite drawing code                          *
; *                        (Scan line version)                          *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************

.386p 
.model flat 
.code 
include Gb.Inc

.data
align dword

SSpriteMain     dd  48 dup (-1)         ; table of sprites in frame.
SSpriteLine     db  512 dup (0)         ; sprite information table (curr line)
SSpriteCount    db  0                   ; sprites on current line.
SSHeight        db  8                   ; sprite height in pixels.
SSMask          db  255                 ; tile number mask

align dword
      
Palette0        dd      0,0,0           ; eax,ecx,edx for palette 0
Palette1        dd      0,0,0           ; likewise for palette 1

TestBit         db      0               ; testing bit for type

.code

ssRecordSize    equ     16              ; size of one record

ssVideo         equ     0               ; Video Ram address (dword), -1=end
ssBitmap        equ     4               ; 16-bit bitmap (word)
ssPalette       equ     6               ; object palette number (byte)
ssType          equ     7               ; sprite behind/in front bit

; ***********************************************************************
; *                                                                     *
; *     Build the table of sprites as 32-bit dwords containing the      *
; *     sprite offset in the low 16 bits, x in bits 24..31              *
; *     and y in 16..23 and sort it                                     *
; *                                                                     *
; ***********************************************************************

SSBuildTable Proc
        push    eax ebx ecx
        mov     eax,0FEA0h              ; first sprite
        mov     ebx,offset SSpriteMain  ; first table entry.
SSBT1:  sub     eax,4
        mov     [ebx],eax
        mov     cl,[byte ptr Memory+eax] ; y at +2
        mov     [byte ptr ebx+2],al
        mov     cl,[byte ptr Memory+eax+1] ; x at +3
        mov     [byte ptr ebx+3],al     ; when we sort it it'll be in x,y order
        add     ebx,4
        cmp     eax,0FE00h              ; last entry ?
        jne     SSBT1
        mov     eax,-1                  ; add end of table marker (-1)
        mov     [ebx],eax
        pop     ecx ebx eax
        ret
EndP SSBuildTable

; ***********************************************************************
; *                                                                     *
; *     Extract all the sprites needed for the current scan line        *
; *     (up to 10) and store the screen address,type,palette and        *
; *     Bitmap in the table                                             *
; *                                                                     *
; ***********************************************************************
           
SSExtractLine Proc
        mov     [SSHeight],8            ; set the correct height
        mov     [SSMask],255
        test    [byte ptr Memory+0FF40h],4
        jz      Not16High
        mov     [SSHeight],16
        mov     [SSMask],254
Not16High:
        mov     esi,offset SSpriteMain  ; come from this table
        mov     edx,offset SSpriteLine  ; where its going
        mov     [SSpriteCount],0        ; clear the sprite counter
SSEL1:  mov     edi,[esi]               ; get the next sprite
        cmp     edi,-1                  ; reached the end ?
        je      SSEOut
        add     esi,4                   ; advance the pointer
        and     edi,0FFFFh              ; get the address bit (lower 16 bits)
        add     edi,offset Memory       ; esi now points to sprite info

        mov     al,[byte ptr Memory+0FF44h] ; al = scan line
        mov     ah,[edi+0]              ; get the Y Coordinate
        sub     ah,16                   ; it starts from -16
        sub     al,ah                   ; al = offset in sprite (vertically)
        cmp     al,[SSHeight]           ; if >= Height          (used later)
        jae     SSEL1                   ; try the next one

        xor     ecx,ecx                 ; ecx = x position - 8
        mov     cl,[edi+1]
        sub     ecx,8
        cmp     ecx,-8                  ; check range
        jle     SSEL1
        cmp     ecx,160
        jge     SSEL1

        mov     bl,[edi+3]              ; check for y flip
        test    bl,040h
        jz      NoYFlipSprite
        mov     bl,[SSHeight]           ; if so al = Height-1-al
        dec     bl
        sub     bl,al
        mov     al,bl
NoYFlipSprite:

        xor     ebx,ebx                 ; work out the display address
        mov     bh,[byte ptr Memory+0FF44h]
        add     ebx,16*256+16           ; this is where it goes.
        add     ebx,offset DisplayImage
        add     ebx,ecx                 ; add the x offset
        mov     [edx+ssVideo],ebx       ; save the write address

        mov     bl,[edi+2]              ; get the pattern number
        and     bl,[SSMask]
        movzx   ebx,bl                  ; into 32 bits
        shl     ebx,4                   ; x 16 (16 bytes per pattern)
        add     ebx,offset Memory+8000h ; add address of sprite tiles.
        movzx   eax,al                  ; zero extend the offset in the tile
        add     eax,eax                 ; 2 bytes per scan line
        add     ebx,eax                 ; get the address of the tile
        mov     bx,[ebx]                ; read the tile pattern.

        mov     cl,[edi+3]              ; check for reverse
        test    cl,020h
        jz      NoSpriteReverse
        movzx   eax,bl
        mov     bl,[byte ptr ReversePixel+eax]
        movzx   eax,bh
        mov     bh,[byte ptr ReversePixel+eax]
NoSpriteReverse:

        mov     [edx+ssBitmap],bx       ; and save it

        mov     al,[edi+3]              ; get palette number bit
        and     al,010h                 ; which palette ?
        mov     [edx+ssPalette],al      ; save the result

        mov     al,[edi+3]              ; get sprite position bit
        and     al,080h
        mov     [edx+ssType],al
    
        add     edx,ssRecordSize        ; next record
        inc     [SSpriteCount]          ; increment the sprite count
        cmp     [SSpriteCount],10
        jne     SSEL1
SSEOut: mov     eax,-1                  ; mark the end of the table
        mov     [edx],eax

        cmp     [SSpriteCount],0        ; if no sprites on this line
        jz      SSEOut2                 ; don't bother creating the palettes

        mov     bl,[byte ptr Memory+0FF48h]
        call    LoadPaletteData         ; create the palettes as usual
        add     eax,08080808h           ; so we don't have to do it for
        add     ecx,08080808h           ; every scan line
        add     edx,08080808h
        mov     [Palette0+0],eax
        mov     [Palette0+4],ecx
        mov     [Palette0+8],edx

        mov     bl,[byte ptr Memory+0FF49h]
        call    LoadPaletteData
        add     eax,0C0C0C0Ch
        add     ecx,0C0C0C0Ch
        add     edx,0C0C0C0Ch
        mov     [Palette1+0],eax
        mov     [Palette1+4],ecx
        mov     [Palette1+8],edx
SSEOut2:
        ret
EndP SSExtractLine

; ***********************************************************************
; *                                                                     *
; *                     Draw all sprites of type AL                     *
; *                                                                     *
; ***********************************************************************

SSDraw Proc
        cmp     [FastSprites],0         ; exit quick if no fast sprites
        jnz     SSDExitQuick

        push    edi esi ebp
        mov     [TestBit],al            ; save the test bit

        mov     edi,offset SSpriteLine  ; pointer to displayable sprites
SSD1:   mov     ebp,[edi]               ; get the first video address
        cmp     ebp,-1                  ; if -1 then quit
        je      SSDExit

        mov     al,[byte ptr edi+ssType]; which type is it ?
        cmp     al,[TestBit]            ; if not that being drawn
        jne     SSDNext                 ; skip it

        mov     bx,[edi+ssBitMap]       ; get the bitmap in bx
        mov     al,[byte ptr edi+ssPalette]
        test    al,16
        jnz     SSDPal1                 ; decide which palette we want.

        mov     eax,[Palette0+0]        ; load in palette 0
        mov     ecx,[Palette0+4]
        mov     edx,[Palette0+8]
        jmp     SSDPalLoaded
SSDPal1:
        mov     eax,[Palette1+0]        ; load in palette 1
        mov     ecx,[Palette1+4]
        mov     edx,[Palette1+8]
SSDPalLoaded:
        call    RenderByte              ; draw the sprite line

SSDNext:
        add     edi,ssRecordSize        ; go to the next one
        jmp     SSD1

SSDExit:pop     ebp esi edi

SSDExitQuick:
        ret
EndP SSDraw

end
